%
% Copyright (c) 2005 Carnegie Mellon University.  All rights
% reserved.
%
% Redistribution and use in source and binary forms, with or without
% modification, are permitted provided that the following conditions
% are met:
%
% 1. Redistributions of source code must retain the above copyright
%    notice, this list of conditions and the following disclaimer.
%
% 2. Redistributions in binary form must reproduce the above copyright
%    notice, this list of conditions and the following disclaimer in
%    the documentation and/or other materials provided with the
%    distribution.
%
% 3. The end-user documentation included with the redistribution,
%    if any, must include the following acknowledgment:
%       "This product includes software developed by
%        Carnegie Mellon University (http://www.cmu.edu/)."
%    Alternately, this acknowledgment may appear in the software itself,
%    if and wherever such third-party acknowledgments normally appear.
%
% 4. The names "Carnegie Mellon University" and "CMU" must
%    not be used to endorse or promote products derived from this
%    software.
%
% 5. Products derived from this software may not be called
%     "Carnegie Mellon University" or "CMU",
%    nor may "Carnegie Mellon University"  or "CMU" appear in their name.
%
% THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
% WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
% OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
% DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
% ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
% SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
% LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
% USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
% ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
% OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
% OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
% SUCH DAMAGE.
%
% @ProposedRating yellow jingcao@ece.cmu.edu
%

function fast_total_buffer(num_of_rows, num_of_cols, total_buffer)
%fast_total_buffer(num_of_rows, num_of_columns, total_buffer)
%Automatically generates a buffer configure file at the end of
%optimization with the default file name 'total_optimal_buffer.config'
%Compared to total_buffer_optimization, fast_total_buffer
%always selects the channel which has the smallest mu_eff

%% you may need to change the following path so that it points to
%% to the correct path of your installation of the .m files. 
path(path, '../../../../../../../util/noctools/matlab_functions/');

out_buffer_config_file = 'fast_total_optimal_buffer.config';

%% WARNING: change 0.25 below if you have different channel service rate.
%% otherwise fsolve may have problem to converge.
mu0 = 1/3 * ones(num_of_rows, num_of_cols, 5);
y = zeros(num_of_rows, num_of_cols, 5);

%% buffer depth to be passed to equations
global b;
b = zeros(num_of_rows, num_of_cols, 5);

global lambda;
equations(mu0);    %% just to have lambda initialized

%% Here we generate starting point of buffer depth. 
%% Basically, all the input buffers with zero load will be
%% initialized with a depth of 0. We don't need to take care 
%% of the local input buffer depth, as they will not be used in the
%% solver anyway.
for i = 1 : num_of_rows
  for j = 1 : num_of_cols
    for k = 2 : 5                 %% no go for local
      if lambda(i,j,k) == 0
        b(i,j,k) = 0;
      else
        b(i,j,k) = 1;
      end
    end
  end
end

%% get the sum of total buffer usage
cur_total_buffer = sum(sum(sum(b)));

if cur_total_buffer > total_buffer
  disp('Total buffer is less than the minimum required buffer');
  return
end


% options=optimset('Display','iter');   % Option to display output
options=optimset('Display','off');      % Don't display

last_mu = mu0;

latency = 10000000;

start_t = cputime;

%% now let's start the optimization
while sum(sum(sum(b))) < total_buffer
  prompt = sprintf('Current total buffer in noc is %d flits with latency indication metric %d', ...
                   sum(sum(sum(b))), latency);
  disp(prompt);

  [mu,info]=fsolve(@equations,last_mu,options);

  %% now calculate the average packet latency by sum(lambda/(mu-lambda))
  latency = calc_latency(lambda, mu);
  
  [x,y,dir] = select_channel(lambda, mu);
  %% increase the depth of next_channel by 1
  b(x,y,dir) = b(x,y,dir) + 1;
  last_mu = mu;
end


end_t = cputime;  

prompt = sprintf('Total used cpu time in this optimization = %f', end_t-start_t);
disp(prompt);

%% let's calculate the latency again here          
[mu,info]=fsolve(@equations,last_mu,options);
latency = calc_latency(lambda,mu);


%% output results 
disp('Write buffer configuration to buffer config file');

fid = fopen(out_buffer_config_file, 'w');
fprintf(fid, '# Buffer configuration file automatically generate using matlab optimization.\n');
fprintf(fid, '# Number of rows: %d, number of columns: %d\n', num_of_rows, num_of_cols);
fprintf(fid, '# Total buffer length: %d\n', total_buffer);
fprintf(fid, '# Optimization CPU time: %f\n', end_t-start_t);

for i = 1 : num_of_rows
    for j = 1 : num_of_cols
      router_id = (i-1) * num_of_cols + (j-1);
      fprintf(fid, '@NODE %d\n', router_id);
      fprintf(fid, 'north_input_channel_buffer_size %d\n', b(i,j, 2));
      fprintf(fid, 'south_input_channel_buffer_size %d\n', b(i,j, 3));
      fprintf(fid, 'east_input_channel_buffer_size %d\n', b(i,j, 4));
      fprintf(fid, 'west_input_channel_buffer_size %d\n', b(i,j, 5));
    end
end

fclose(fid);

%% verify results by checking all the non-zero load buffers have
%% non-zero depth
for i = 1 : num_of_rows
    for j = 1 : num_of_cols
      for k = 2 : 5
        if lambda(i,j,k) > 0 & b(i,j,k) == 0
          disp('Incorrect solution with wrong zero buffer depth assignment');
        end
      end
    end
end

