/*
 * Copyright (c) 2005 Carnegie Mellon University.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by
 *        Carnegie Mellon University (http://www.cmu.edu/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Carnegie Mellon University" and "CMU" must
 *    not be used to endorse or promote products derived from this
 *    software.
 *
 * 5. Products derived from this software may not be called
 *     "Carnegie Mellon University" or "CMU",
 *    nor may "Carnegie Mellon University"  or "CMU" appear in their name.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * @ProposedRating yellow jingcao@ece.cmu.edu
 */

#include "arch.hpp"
#include "gVal.hpp"
#include "util.hpp"

Tile::Router::Router() {
    routing_table = new int*[gTileNum];
    for(int i=0; i<gTileNum; i++) 
        routing_table[i] = new int[gTileNum];
    for(int i=0; i<gTileNum; i++) 
        for(int j=0; j<gTileNum; j++) 
            routing_table[i][j] = -2;
}

Tile::Router::~Router() {
    if(routing_table) {
        for(int i=0; i<gTileNum; i++) 
            delete []routing_table[i];
        delete []routing_table;
    }
}

bool Tile::Router::initialize(ROUTING_TYPE r_type, pTile host) {
    host_tile = host;
    routing_type = r_type;
    if (routing_type == routing_xy) 
        return generate_xy_routing_table();
    else {
        cerr<<"This architecture and routing combination is not supported."<<endl;
        return false;
    }
}

//This method is used to generate the fixed, non-adaptive, minimal routing 
//table. This method is only applicable to Mesh or Torus
bool Tile::Router::generate_xy_routing_table() {
    for(int i=0; i<totalTiles; i++) 
        for(int j=0; j<totalTiles; j++) 
            routing_table[i][j] = -2;
  
    for(int dstTile=0; dstTile<gTileNum; dstTile++) {
        if(dstTile == host_tile->id) {                     //deliver to me
            routing_table[0][dstTile] = -1;
            continue;
        }

        //check out the dst Tile's position first
        Position dstPos;
        dstPos.row = dstTile / g_edge_size;
        dstPos.col = dstTile % g_edge_size;

        Position nextStep = host_tile->pos;

        if(dstPos.col != host_tile->pos.col) {            //We should go horizontally
            if(host_tile->pos.col > dstPos.col) 
                nextStep.col--;
            else 
                nextStep.col++;
        }
        else {            //We should go vertically
#ifdef DEBUG
            //Sanity checking
            if(dstPos.row == host_tile->pos.row) {
                cerr << "Routing table creation error. " << endl;
                return false;
            }
#endif //DEBUG
            if(host_tile->pos.row > dstPos.row) 
                nextStep.row--;
            else 
                nextStep.row++;
        }

        int i=0;
        for(i=0; i<host_tile->goLinkNum; i++) {
            pLink pL = host_tile->goLink[i];
            if(pL->ToTilePos() == nextStep) {
                routing_table[0][dstTile] = pL->GetId();
                break;
            }
        }
#ifdef DEBUG
        //Sanity checking
        if(i==host_tile->goLinkNum) {
            cerr << "Routing table creation error." << endl;
            return false;
        }
#endif //DEBUG
    }

    //Duplicate this routing row to the other routing rows.
    for(int i=1; i<gTileNum; i++) 
        for(int j=0; j<gTileNum; j++) 
            routing_table[i][j] = routing_table[0][j];

    return true;
}

int Tile::Router::route_to_link(int src_tile, int dst_tile) const {
    return routing_table[src_tile][dst_tile];
}

int Tile::Router::set_routing_entry(int src_tile, int dst_tile, int link_id) {
    routing_table[src_tile][dst_tile] = link_id;
    return link_id;
}

int Tile::Router::choose_egress(int src_tile, int dst_tile) {
    return routing_table[src_tile][dst_tile];
}
